var assert = require('assert');
var slug   = require('../index.js');

describe('Laravel Slug', function () {

    describe('Delimiters', function () {
        it('should be placed instead of spaces', function () {
            var stringWithSpaces = 'this is a string';
            assert.equal(slug(stringWithSpaces), 'this-is-a-string');
        });
    });

    describe('Trimming', function () {
        it('should happen to strings surrounded by whitespace', function () {
            var paddedString = '   these are words   ';
            assert.equal(slug(paddedString), 'these-are-words');
        });
    });

    describe('Transliteration', function () {
        it('should transliterate special alphabets, the same way Laravel does', function () {
            var russianString = 'это текст';
            assert.equal(slug(russianString), 'eto-tekst');
        });
    });

    describe('Other Character Dashing', function () {
        it('should replace underscores, question marks, and dashes with dashes', function () {
            var funnyString = 'what_is-this even? A string';
            assert.equal(slug(funnyString), 'what-is-this-even-a-string');
        });
    });

    describe('Non Alphanumeric Character Removal', function () {
        it('Removes anything that is\'t a letter, number, whitespace or separator', function () {
            var stringWithChars = 'these [are brackets]';
            assert.equal(slug(stringWithChars), 'these-are-brackets');
        });
    })
});
